import logging
from pathlib import Path
from time import time, sleep
from typing import Optional

import pyglet

from ververser.asset_manager import AssetManager, ReloadStatus
from ververser.fps_counter import FPSCounter
from ververser.main_script import MainScript
from ververser.global_game_window import set_global_game_window


logger = logging.getLogger(__name__)


class GameWindow( pyglet.window.Window ):

    def __init__( self, asset_folder_path : Path, throttle_fps = 30, make_global = True ):
        super().__init__(vsync = False)

        self.throttle_fps = throttle_fps

        self.alive = True

        self.frame_count = 0
        self.last_update = time()
        self.fps_counter = FPSCounter()

        self.is_paused = False
        self.has_asset_problem = False

        self.asset_manager = AssetManager( asset_folder_path, self )
        self.main_script : Optional[ MainScript ] = None

        if make_global:
            set_global_game_window( self )

    def on_close(self):
        self.alive = False

    def run(self):
        while self.alive:
            self.dispatch_events()

            if self.is_paused:
                continue

            now = time()
            dt = now - self.last_update
            self.last_update = now

            if self.throttle_fps:
                sleep_time = ( 1 / self.throttle_fps ) - dt
                sleep( max( sleep_time, 0 ) )

            # TODO:
            # we do not want the framerate to affect physics
            # easiest way to do that is fix the dt here for now
            dt = 1/60

            reload_status = self.asset_manager.try_reload()
            if reload_status == ReloadStatus.FAILED:
                logger.info( "Error occured during asset loading. Game is now paused!" )
                self.has_asset_problem = True
                continue
            else:
                if reload_status == ReloadStatus.RELOADED:
                    self.has_asset_problem = False

            if self.is_paused or self.has_asset_problem:
                continue

            self._update(dt)
            self.update(dt)

            self._draw_start()
            self.draw()
            self._draw_end()

    def _update( self, dt ):
        self.fps_counter.update()

    def _draw_start( self ):
        self.clear()

    def _draw_end( self ):
        self.fps_counter.draw()
        self.flip()
        self.frame_count += 1

    # ---------------- Convenience Functions ----------------
    def try_invoke( self, f ):
        try :
            f()
        except :
            logging.exception( "Error occured during script invokation. Game is now paused!" )
            self.has_asset_problem = True

    # ================ End of standard boilerplate ================
    # ================ Overload the methods below! ================

    def init( self ):
        self.main_script = self.asset_manager.load_main_script()

    def update( self, dt ):
        self.try_invoke( lambda : self.main_script.vvs_update( dt ) )

    def draw( self ):
        self.try_invoke( lambda : self.main_script.vvs_draw() )


import logging
from pathlib import Path
from ververser import GameWindow


if __name__ == '__main__':
    logging.basicConfig( level = logging.INFO )
    game = GameWindow( Path( __file__ ).parent / 'content' )
    game.run()



from pathlib import Path
from typing import Any, Callable

from ververser.reloading_asset import ReloadingAsset, ReloadStatus
from ververser.main_script import load_main_script, MainScript
from ververser.folder_watcher import FolderWatcher
from ververser.script import load_script


AssetLoaderType = Callable[ [ Path ], Any ]

EXPECTED_MAIN_SCRIPT_NAME = Path( 'main.py' )


class AssetManager:

    def __init__( self, asset_folder_path : Path, game_window ):
        self.asset_folder_path = asset_folder_path
        self.game_window = game_window

        self.script_watcher = FolderWatcher( asset_folder_path, '.py' )

        self.asset_loaders : list[ tuple[ str, AssetLoaderType ] ] = []
        self.assets : list[ ReloadingAsset ] = []

        self.register_asset_loader( '.py', load_script )

    def make_asset_path_complete( self, asset_path : Path ) -> Path:
        return self.asset_folder_path / asset_path

    def register_asset_loader( self, postfix : str, f_load_asset : AssetLoaderType ) -> None:
        self.asset_loaders.append( ( postfix, f_load_asset ) )

    def load_main_script( self ) -> MainScript:
        absolute_asset_path = self.make_asset_path_complete( EXPECTED_MAIN_SCRIPT_NAME )
        assert self.exists( absolute_asset_path ), f'Could not load asset. File path: "{EXPECTED_MAIN_SCRIPT_NAME}"'
        main_script = load_main_script( absolute_asset_path, self.game_window )
        return main_script

    def get_asset_loader_for_file( self, file_path : Path ) -> AssetLoaderType:
        # reverse search through all registered loaders
        # this way. newest registered loaders overrule older ones
        for postfix, asset_loader in reversed( self.asset_loaders ):
            if str( file_path ).endswith( postfix ):
                return asset_loader
        assert False, f'No asset loader found for file_path: "{file_path}". Known loaders: {self.asset_loaders}'

    def try_reload( self ) -> ReloadStatus:
        # first check if any script file is outdated
        if self.script_watcher.is_folder_updated():
            self.game_window.init()

        # if scripts are not outdated, also check the other types of assets
        overall_reload_status = ReloadStatus.NOT_CHANGED
        for reloading_asset in self.assets:
            reloading_asset.try_reload()
            reload_status = reloading_asset.reload_status
            if reload_status == ReloadStatus.RELOADED:
                overall_reload_status = ReloadStatus.RELOADED
            if reload_status == ReloadStatus.FAILED:
                return ReloadStatus.FAILED
        return overall_reload_status

    def exists( self, asset_path ) -> bool:
        complete_asset_path = self.make_asset_path_complete( asset_path )
        return complete_asset_path.is_file()

    def load( self, asset_path ) -> ReloadingAsset:
        absolute_asset_path = self.make_asset_path_complete( asset_path.strip() )
        assert self.exists( absolute_asset_path ), f'Could not load asset. File path: "{asset_path}"'
        asset_loader = self.get_asset_loader_for_file( absolute_asset_path )
        reloading_asset = ReloadingAsset(
            f_load_asset = lambda path : asset_loader( path ),
            file_path = absolute_asset_path
        )
        self.assets.append( reloading_asset )
        reloading_asset.try_reload()
        return reloading_asset



from enum import Enum, auto
import logging
from typing import Any

from ververser.file_watcher import FileWatcher


class ReloadStatus(Enum):
    NOT_CHANGED = auto()
    RELOADED = auto()
    FAILED = auto()


class ReloadingAsset:

    def __init__(self, f_load_asset, file_path):
        self.f_load_asset = f_load_asset
        self.file_watcher = FileWatcher(file_path)

    def __getattr__( self, name : str ) -> Any:
        return getattr( self.get(), name )

    def try_reload( self ) -> None:
        if not self.file_watcher.is_file_updated():
            self.reload_status = ReloadStatus.NOT_CHANGED
            return
        asset_path = self.file_watcher.file_path
        try:
            self.asset = self.f_load_asset( asset_path )
        except Exception as e :
            logging.error( f'Encountered an error during loading of asset from file "{asset_path}"' )
            logging.exception( e )
            self.reload_status = ReloadStatus.FAILED
            return
        self.reload_status =  ReloadStatus.RELOADED
        return

    def get( self ) -> Any:
        return self.asset





import os.path
from pathlib import Path


class FolderWatcher:

    def __init__( self, folder_path: Path, postfix : str ):
        self.folder_path = folder_path
        self.postfix = postfix
        self.last_seen_time_modified = None

    def get_last_time_modified( self ):
        all_file_paths = self.folder_path.rglob(f'*{self.postfix}')
        all_modification_times = [ os.path.getmtime( file_path ) for file_path in all_file_paths ]
        modification_time = max( all_modification_times )
        return modification_time

    def is_folder_updated( self ) -> bool:
        last_time_modified = self.get_last_time_modified()
        is_updated = False
        if last_time_modified != self.last_seen_time_modified:
            is_updated = True
            self.last_seen_time_modified = last_time_modified
            print(f'Folder was updated for extension {self.postfix}! - {self.folder_path} - Timestamp: {self.last_seen_time_modified}')
        return is_updated



from .asset_manager import AssetManager, AssetLoaderType
from .file_watcher import FileWatcher
from .folder_watcher import FolderWatcher
from .fps_counter import FPSCounter
from .game_window import GameWindow
from .reloading_asset import ReloadingAsset, ReloadStatus
from .script import import_script, load_script, Script
from .global_game_window import set_global_game_window, get_global_game_window
